# PowerModels Experiment Results

This section presents results of running PowerModels on collections of
established power network test cases from the [PGLib-OPF](https://arxiv.org/abs/1908.02788) archive.
This provides validation of PowerModels as well as a results baseline for these test cases.
All models were solved using [IPOPT](https://link.springer.com/article/10.1007/s10107-004-0559-y).



## Experiment Design
This experiment consists of running the following PowerModels commands,
```
solver = optimizer_with_attributes(Ipopt.Optimizer, "tol" => 1e-6)
result_ac  = run_opf(case,   ACPPowerModel, solver)
result_soc = run_opf(case, SOCWRPowerModel, solver)
result_qc  = run_opf(case,  QCRMPowerModel, solver)
```
for each case in the PGLib-OPF archive.
If the value of `result["termination_status"]` is `LOCALLY_SOLVED` then the
values of `result["objective"]` and `result["solve_time"]` are reported,
otherwise an `err.` or `--` is displayed.  A value of `n.d.` indicates that no data was available.
  The optimality gap is defined as,
```
soc_gap = 100*(result_ac["objective"] - result_soc["objective"])/result_ac["objective"]
```

It is important to note that the `result["solve_time"]`
value in this experiment does not include Julia's JIT time, about 2-5 seconds.
The results were computed using the [HSL](http://www.hsl.rl.ac.uk/ipopt/) ma57 solver in IPOPT.
The default linear solver provided with Ipopt.jl will increase the runtime by 2-6x.


## Software Versions
**[PowerModels.jl](https://github.com/lanl-ansi/PowerModels.jl):** v0.17

**[Ipopt.jl](https://github.com/JuliaOpt/Ipopt.jl):** v0.6

**[PGLib-OPF](https://github.com/power-grid-lib/pglib-opf):** v19.05

**Hardware:** Dual Intel 2.10GHz CPUs, 128GB RAM


## Typical Operating Conditions (TYP)
| **Case Name** | **Nodes** | **Edges** | **AC (\$/h)** | **QC Gap (%)** | **SOC Gap (%)** | **AC Time (sec.)** | **QC Time (sec.)** | **SOC Time (sec.)** |
| ------------- | --------- | --------- | ------------- | -------------- | --------------- | ------------------ | ------------------ | ------------------- |
| `case3_lmbd` | 3 | 3 | 5.8126e+03 | 1.22 | 1.32 | <1 | <1 | <1 |
| `case5_pjm` | 5 | 6 | 1.7552e+04 | 14.55 | 14.55 | <1 | <1 | <1 |
| `case14_ieee` | 14 | 20 | 2.1781e+03 | 0.11 | 0.11 | <1 | <1 | <1 |
| `case24_ieee_rts` | 24 | 38 | 6.3352e+04 | 0.02 | 0.02 | <1 | <1 | <1 |
| `case30_as` | 30 | 41 | 8.0313e+02 | 0.06 | 0.06 | <1 | <1 | <1 |
| `case30_fsr` | 30 | 41 | 5.7577e+02 | 0.39 | 0.39 | <1 | <1 | <1 |
| `case30_ieee` | 30 | 41 | 8.2085e+03 | 18.81 | 18.84 | <1 | <1 | <1 |
| `case39_epri` | 39 | 46 | 1.3842e+05 | 0.55 | 0.56 | <1 | <1 | <1 |
| `case57_ieee` | 57 | 80 | 3.7589e+04 | 0.16 | 0.16 | <1 | <1 | <1 |
| `case73_ieee_rts` | 73 | 120 | 1.8976e+05 | 0.04 | 0.04 | <1 | <1 | <1 |
| `case89_pegase` | 89 | 210 | 1.0729e+05 | 0.75 | 0.75 | <1 | <1 | <1 |
| `case118_ieee` | 118 | 186 | 9.7214e+04 | 0.79 | 0.91 | <1 | <1 | <1 |
| `case162_ieee_dtc` | 162 | 284 | 1.0808e+05 | 5.84 | 5.95 | <1 | <1 | <1 |
| `case179_goc` | 179 | 263 | 7.5427e+05 | 0.16 | 0.16 | <1 | <1 | <1 |
| `case200_tamu` | 200 | 245 | 2.7558e+04 | 0.01 | 0.01 | <1 | <1 | <1 |
| `case240_pserc` | 240 | 448 | 3.3297e+06 | 2.73 | 2.78 | 3 | 4 | 2 |
| `case300_ieee` | 300 | 411 | 5.6522e+05 | 2.58 | 2.63 | <1 | <1 | <1 |
| `case500_tamu` | 500 | 597 | 7.2578e+04 | 5.39 | 5.39 | <1 | 2 | <1 |
| `case588_sdet` | 588 | 686 | 3.1314e+05 | 1.91 | 2.14 | <1 | 2 | <1 |
| `case1354_pegase` | 1354 | 1991 | 1.2588e+06 | 1.56 | 1.57 | 5 | 6 | 3 |
| `case1888_rte` | 1888 | 2531 | 1.4025e+06 | 2.05 | 2.05 | 9 | 8 | 31 |
| `case1951_rte` | 1951 | 2596 | 2.0856e+06 | 0.13 | 0.14 | 19 | 9 | 5 |
| `case2000_tamu` | 2000 | 3206 | 1.2285e+06 | 0.20 | 0.21 | 10 | 9 | 3 |
| `case2316_sdet` | 2316 | 3017 | 1.7753e+06 | 1.79 | 1.80 | 6 | 11 | 4 |
| `case2383wp_k` | 2383 | 2896 | 1.8682e+06 | 0.97 | 1.04 | 7 | 10 | 5 |
| `case2736sp_k` | 2736 | 3504 | 1.3080e+06 | 0.30 | 0.31 | 6 | 10 | 4 |
| `case2737sop_k` | 2737 | 3506 | 7.7773e+05 | 0.26 | 0.27 | 5 | 8 | 3 |
| `case2746wop_k` | 2746 | 3514 | 1.2083e+06 | 0.36 | 0.37 | 5 | 9 | 4 |
| `case2746wp_k` | 2746 | 3514 | 1.6317e+06 | 0.32 | 0.33 | 6 | 9 | 4 |
| `case2848_rte` | 2848 | 3776 | 1.2866e+06 | 0.12 | 0.13 | 17 | 13 | 7 |
| `case2853_sdet` | 2853 | 3921 | 2.0524e+06 | 0.87 | 0.91 | 9 | 12 | 6 |
| `case2868_rte` | 2868 | 3808 | 2.0096e+06 | 0.10 | 0.10 | 15 | 14 | 8 |
| `case2869_pegase` | 2869 | 4582 | 2.4628e+06 | 1.01 | 1.01 | 11 | 20 | 7 |
| `case3012wp_k` | 3012 | 3572 | 2.6008e+06 | 0.98 | 1.03 | 9 | 14 | 13 |
| `case3120sp_k` | 3120 | 3693 | 2.1480e+06 | 0.55 | 0.56 | 9 | 12 | 5 |
| `case3375wp_k` | 3374 | 4161 | 7.4382e+06 | 0.54 | 0.55 | 11 | 29 | 6 |
| `case4661_sdet` | 4661 | 5997 | 2.2513e+06 | 1.89 | 1.99 | 15 | 27 | 12 |
| `case6468_rte` | 6468 | 9000 | 2.0697e+06 | 1.12 | 1.13 | 64 | 87 | 27 |
| `case6470_rte` | 6470 | 9005 | 2.2376e+06 | 1.75 | 1.76 | 36 | 43 | 24 |
| `case6495_rte` | 6495 | 9019 | 3.0678e+06 | 15.09 | 15.11 | 69 | 65 | 25 |
| `case6515_rte` | 6515 | 9037 | 2.8255e+06 | 6.39 | 6.40 | 53 | 44 | 23 |
| `case9241_pegase` | 9241 | 16049 | 6.2431e+06 | 1.71 | 2.54 | 49 | 104 | 36 |
| `case10000_tamu` | 10000 | 12706 | 2.4859e+06 | 0.37 | 0.72 | 98 | 67 | 33 |
| `case13659_pegase` | 13659 | 20467 | 8.9480e+06 | 0.98 | 1.39 | 59 | 121 | 62 |


## Congested Operating Conditions (API)
| **Case Name** | **Nodes** | **Edges** | **AC (\$/h)** | **QC Gap (%)** | **SOC Gap (%)** | **AC Time (sec.)** | **QC Time (sec.)** | **SOC Time (sec.)** |
| ------------- | --------- | --------- | ------------- | -------------- | --------------- | ------------------ | ------------------ | ------------------- |
| `case3_lmbd__api` | 3 | 3 | 1.1242e+04 | 5.63 | 9.32 | <1 | <1 | <1 |
| `case5_pjm__api` | 5 | 6 | 7.6377e+04 | 4.09 | 4.09 | <1 | <1 | <1 |
| `case14_ieee__api` | 14 | 20 | 5.9994e+03 | 5.13 | 5.13 | <1 | <1 | <1 |
| `case24_ieee_rts__api` | 24 | 38 | 1.3495e+05 | 13.01 | 17.87 | <1 | <1 | <1 |
| `case30_as__api` | 30 | 41 | 4.9962e+03 | 44.61 | 44.61 | <1 | <1 | <1 |
| `case30_fsr__api` | 30 | 41 | 7.0115e+02 | 2.76 | 2.76 | <1 | <1 | <1 |
| `case30_ieee__api` | 30 | 41 | 1.8044e+04 | 5.46 | 5.46 | <1 | <1 | <1 |
| `case39_epri__api` | 39 | 46 | 2.4975e+05 | 1.72 | 1.74 | <1 | <1 | <1 |
| `case57_ieee__api` | 57 | 80 | 4.9297e+04 | 0.09 | 0.09 | <1 | <1 | <1 |
| `case73_ieee_rts__api` | 73 | 120 | 4.2273e+05 | 11.07 | 12.89 | <1 | <1 | <1 |
| `case89_pegase__api` | 89 | 210 | 1.3428e+05 | 13.45 | 13.47 | <1 | <1 | <1 |
| `case118_ieee__api` | 118 | 186 | 2.4205e+05 | 28.70 | 28.81 | <1 | <1 | <1 |
| `case162_ieee_dtc__api` | 162 | 284 | 1.2100e+05 | 4.33 | 4.36 | <1 | <1 | <1 |
| `case179_goc__api` | 179 | 263 | 1.9321e+06 | 5.93 | 9.88 | <1 | 2 | <1 |
| `case200_tamu__api` | 200 | 245 | 3.6763e+04 | 0.03 | 0.03 | <1 | <1 | <1 |
| `case240_pserc__api` | 240 | 448 | 4.7681e+06 | 0.69 | 0.74 | 4 | 5 | 2 |
| `case300_ieee__api` | 300 | 411 | 6.5015e+05 | 0.84 | 0.89 | <1 | <1 | <1 |
| `case500_tamu__api` | 500 | 597 | 4.2776e+04 | 2.92 | 2.92 | <1 | 2 | <1 |
| `case588_sdet__api` | 588 | 686 | 3.9476e+05 | 1.39 | 1.61 | <1 | 2 | <1 |
| `case1354_pegase__api` | 1354 | 1991 | 1.4867e+06 | 0.65 | 0.66 | 5 | 7 | 3 |
| `case1888_rte__api` | 1888 | 2531 | 1.9539e+06 | 0.22 | 0.23 | 8 | 15 | 6 |
| `case1951_rte__api` | 1951 | 2596 | 2.4168e+06 | 0.53 | 0.55 | 9 | 9 | 5 |
| `case2000_tamu__api` | 2000 | 3206 | 1.2850e+06 | 2.66 | 2.71 | 13 | 10 | 3 |
| `case2316_sdet__api` | 2316 | 3017 | 2.1890e+06 | 1.79 | 1.80 | 8 | 11 | 4 |
| `case2383wp_k__api` | 2383 | 2896 | 2.7913e+05 | 0.01 | 0.01 | 2 | 3 | <1 |
| `case2736sp_k__api` | 2736 | 3504 | 6.5394e+05 | 10.83 | 10.84 | 7 | 9 | 3 |
| `case2737sop_k__api` | 2737 | 3506 | 3.6920e+05 | 6.41 | 6.41 | 6 | 4 | 2 |
| `case2746wop_k__api` | 2746 | 3514 | 5.1166e+05 | 0.01 | 0.01 | 2 | 3 | 2 |
| `case2746wp_k__api` | 2746 | 3514 | 5.8183e+05 | 0.01 | 0.00 | 3 | 5 | 2 |
| `case2848_rte__api` | 2848 | 3776 | 1.4760e+06 | 0.25 | 0.25 | 28 | 13 | 6 |
| `case2853_sdet__api` | 2853 | 3921 | 2.4579e+06 | 1.92 | 1.96 | 10 | 13 | 6 |
| `case2868_rte__api` | 2868 | 3808 | 2.3405e+06 | 0.18 | 0.19 | 26 | 15 | 6 |
| `case2869_pegase__api` | 2869 | 4582 | 2.9342e+06 | 1.32 | 1.33 | 13 | 19 | 8 |
| `case3012wp_k__api` | 3012 | 3572 | 7.2887e+05 | 0.00 | 0.00 | 4 | 5 | 2 |
| `case3120sp_k__api` | 3120 | 3693 | 9.8428e+05 | 23.54 | 23.60 | 12 | 13 | 4 |
| `case3375wp_k__api` | 3374 | 4161 | 5.8478e+06 | 9.35 | -- | 11 | 22 | 347 |
| `case4661_sdet__api` | 4661 | 5997 | 2.6951e+06 | 2.54 | 2.64 | 17 | 27 | 50 |
| `case6468_rte__api` | 6468 | 9000 | 2.3135e+06 | 0.80 | 0.82 | 76 | 55 | 190 |
| `case6470_rte__api` | 6470 | 9005 | 2.6065e+06 | 1.19 | 1.20 | 55 | 38 | 68 |
| `case6495_rte__api` | 6495 | 9019 | 3.0265e+06 | 2.97 | 3.01 | 55 | 46 | 23 |
| `case6515_rte__api` | 6515 | 9037 | 3.1363e+06 | 1.91 | 1.95 | 63 | 46 | 22 |
| `case9241_pegase__api` | 9241 | 16049 | 7.0264e+06 | 1.80 | -- | 118 | 262 | 1617 |
| `case10000_tamu__api` | 10000 | 12706 | 1.8713e+06 | 5.53 | 6.18 | 108 | 117 | 34 |
| `case13659_pegase__api` | 13659 | 20467 | 9.2971e+06 | 1.18 | 1.83 | 65 | 164 | 76 |


## Small Angle Difference Conditions (SAD)
| **Case Name** | **Nodes** | **Edges** | **AC (\$/h)** | **QC Gap (%)** | **SOC Gap (%)** | **AC Time (sec.)** | **QC Time (sec.)** | **SOC Time (sec.)** |
| ------------- | --------- | --------- | ------------- | -------------- | --------------- | ------------------ | ------------------ | ------------------- |
| `case3_lmbd__sad` | 3 | 3 | 5.9593e+03 | 1.42 | 3.75 | <1 | <1 | <1 |
| `case5_pjm__sad` | 5 | 6 | 2.6115e+04 | 0.99 | 3.62 | <1 | <1 | <1 |
| `case14_ieee__sad` | 14 | 20 | 2.7773e+03 | 21.50 | 21.54 | <1 | <1 | <1 |
| `case24_ieee_rts__sad` | 24 | 38 | 7.6943e+04 | 2.93 | 9.56 | <1 | <1 | <1 |
| `case30_as__sad` | 30 | 41 | 8.9749e+02 | 2.32 | 7.88 | <1 | <1 | <1 |
| `case30_fsr__sad` | 30 | 41 | 5.7679e+02 | 0.41 | 0.47 | <1 | <1 | <1 |
| `case30_ieee__sad` | 30 | 41 | 8.2085e+03 | 5.93 | 9.70 | <1 | <1 | <1 |
| `case39_epri__sad` | 39 | 46 | 1.4835e+05 | 0.21 | 0.66 | <1 | <1 | <1 |
| `case57_ieee__sad` | 57 | 80 | 3.8664e+04 | 0.35 | 0.71 | <1 | <1 | <1 |
| `case73_ieee_rts__sad` | 73 | 120 | 2.2775e+05 | 2.54 | 6.75 | <1 | <1 | <1 |
| `case89_pegase__sad` | 89 | 210 | 1.0729e+05 | 0.71 | 0.73 | <1 | <1 | <1 |
| `case118_ieee__sad` | 118 | 186 | 1.0522e+05 | 6.84 | 8.22 | <1 | <1 | <1 |
| `case162_ieee_dtc__sad` | 162 | 284 | 1.0870e+05 | 6.25 | 6.48 | <1 | <1 | <1 |
| `case179_goc__sad` | 179 | 263 | 7.6254e+05 | 1.01 | 1.12 | <1 | <1 | <1 |
| `case200_tamu__sad` | 200 | 245 | 2.7558e+04 | 0.01 | 0.01 | <1 | <1 | <1 |
| `case240_pserc__sad` | 240 | 448 | 3.4071e+06 | 4.41 | 4.98 | 4 | 4 | 2 |
| `case300_ieee__sad` | 300 | 411 | 5.6571e+05 | 2.43 | 2.61 | <1 | <1 | <1 |
| `case500_tamu__sad` | 500 | 597 | 7.9234e+04 | 7.90 | 7.92 | <1 | 2 | <1 |
| `case588_sdet__sad` | 588 | 686 | 3.2986e+05 | 6.10 | 6.81 | <1 | 2 | <1 |
| `case1354_pegase__sad` | 1354 | 1991 | 1.2588e+06 | 1.53 | 1.57 | 5 | 6 | 3 |
| `case1888_rte__sad` | 1888 | 2531 | 1.4139e+06 | 2.81 | 2.82 | 9 | 8 | 24 |
| `case1951_rte__sad` | 1951 | 2596 | 2.0928e+06 | 0.44 | 0.48 | 17 | 9 | 5 |
| `case2000_tamu__sad` | 2000 | 3206 | 1.2303e+06 | 0.33 | 0.35 | 10 | 9 | 3 |
| `case2316_sdet__sad` | 2316 | 3017 | 1.7753e+06 | 1.76 | 1.80 | 6 | 10 | 5 |
| `case2383wp_k__sad` | 2383 | 2896 | 1.9127e+06 | 1.98 | 2.93 | 8 | 10 | 5 |
| `case2736sp_k__sad` | 2736 | 3504 | 1.3273e+06 | 1.36 | 1.63 | 8 | 10 | 5 |
| `case2737sop_k__sad` | 2737 | 3506 | 7.9153e+05 | 1.76 | 1.95 | 7 | 9 | 4 |
| `case2746wop_k__sad` | 2746 | 3514 | 1.2343e+06 | 2.01 | 2.37 | 7 | 8 | 4 |
| `case2746wp_k__sad` | 2746 | 3514 | 1.6676e+06 | 1.67 | 2.22 | 8 | 9 | 5 |
| `case2848_rte__sad` | 2848 | 3776 | 1.2890e+06 | 0.24 | 0.26 | 17 | 13 | 6 |
| `case2853_sdet__sad` | 2853 | 3921 | 2.0701e+06 | 1.69 | 1.74 | 9 | 11 | 6 |
| `case2868_rte__sad` | 2868 | 3808 | 2.0224e+06 | 0.61 | 0.64 | 18 | 13 | 6 |
| `case2869_pegase__sad` | 2869 | 4582 | 2.4689e+06 | 1.02 | 1.13 | 11 | 20 | 7 |
| `case3012wp_k__sad` | 3012 | 3572 | 2.6213e+06 | 1.41 | 1.62 | 10 | 14 | 6 |
| `case3120sp_k__sad` | 3120 | 3693 | 2.1782e+06 | 1.51 | 1.65 | 11 | 14 | 6 |
| `case3375wp_k__sad` | 3374 | 4161 | 7.4382e+06 | 0.50 | 0.55 | 11 | 18 | 6 |
| `case4661_sdet__sad` | 4661 | 5997 | 2.2610e+06 | 1.79 | 1.96 | 16 | 27 | 12 |
| `case6468_rte__sad` | 6468 | 9000 | 2.0697e+06 | 1.10 | 1.12 | 63 | 75 | 31 |
| `case6470_rte__sad` | 6470 | 9005 | 2.2416e+06 | 1.87 | 1.91 | 37 | 42 | 22 |
| `case6495_rte__sad` | 6495 | 9019 | 3.0678e+06 | 14.99 | 15.11 | 70 | 56 | 25 |
| `case6515_rte__sad` | 6515 | 9037 | 2.8826e+06 | 8.22 | 8.26 | 61 | 46 | 22 |
| `case9241_pegase__sad` | 9241 | 16049 | 6.3195e+06 | 2.42 | 2.48 | 59 | 91 | 35 |
| `case10000_tamu__sad` | 10000 | 12706 | 2.4859e+06 | 0.36 | 0.72 | 59 | 68 | 36 |
| `case13659_pegase__sad` | 13659 | 20467 | 9.0433e+06 | 1.64 | 1.69 | 57 | 125 | 43 |


